from seq2science.util import sieve_bam


localrules: setup_blacklist, complement_blacklist


rule setup_blacklist:
    """
    Combine the encode blacklist with mitochrondrial dna depending on config.
    """
    input:
        blacklist=expand("{genome_dir}/{{assembly}}/{{assembly}}.blacklist.bed", **config),
        sizes=expand("{genome_dir}/{{assembly}}/{{assembly}}.fa.sizes", **config),
    output:
        temp(expand("{genome_dir}/{{assembly}}/{{assembly}}.customblacklist.bed", **config)),
    params:
        config.get("remove_blacklist"),  # helps resolve changed params
        config.get("remove_mito"),  # helps resolve changed params
    run:
        newblacklist = ""
        if config.get("remove_blacklist"):
            with open(input["blacklist"][0]) as file:
                newblacklist += file.read()

        if config.get("remove_mito"):
            with open(input["sizes"][0]) as file:
                sizesfile = file.read().strip()
                for match in re.findall("chrM.*|chrm.*|MT.*", sizesfile):
                    chrm, size = match.split("\t")
                    newblacklist += f"{chrm}\t0\t{size}\n"

        with open(output[0], "w") as f:
            f.write(newblacklist)


rule complement_blacklist:
    """
    Take the complement of the blacklist. We need this complement to tell samtools
    to only keep reads that are in the complement of the blacklist.
    """
    input:
        blacklist=rules.setup_blacklist.output,
        sizes=expand("{genome_dir}/{{assembly}}/{{assembly}}.fa.sizes", **config),
    output:
        expand("{genome_dir}/{{assembly}}/{{assembly}}.customblacklist_complement.bed", **config),
    params:
        config.get("remove_blacklist"),  # helps resolve changed params
        config.get("remove_mito"),  # helps resolve changed params
    log:
        expand("{log_dir}/complement_blacklist/{{assembly}}.log", **config),
    benchmark:
        expand("{benchmark_dir}/complement_blacklist/{{assembly}}.benchmark.txt", **config)[0]
    conda:
        "../envs/bedtools.yaml"
    shell:
        """
        sortBed -faidx {input.sizes} -i {input.blacklist} 2>> {log} |
        complementBed -i stdin -g {input.sizes} > {output} 2>> {log}
        """


if config["filter_on_size"]:
    sieve_bam_output = {"allsizes": temp(expand("{result_dir}/{aligner}/{{assembly}}-{{sample}}_allsizes.samtools-coordinate-sieved.sam",**config)),
                        "final": temp(expand("{result_dir}/{aligner}/{{assembly}}-{{sample}}.samtools-coordinate-sieved.bam",**config))}
else:
    sieve_bam_output = {"final": temp(expand("{result_dir}/{aligner}/{{assembly}}-{{sample}}.samtools-coordinate-sieved.bam",**config))}


rule sieve_bam:
    """
    "Sieve" a bam.

    This rule does (at least) one of these:
        * filtering on minimum mapping quality
        * tn5 shift adjustment
        * remove multimappers
        * remove reads inside the blacklist
        * filter paired-end reads on transcript length
     
    """
    input:
        bam=rules.samtools_presort.output,
        blacklist=rules.complement_blacklist.output,
        sizes=expand("{genome_dir}/{{assembly}}/{{assembly}}.fa.sizes", **config),
    output:
        **sieve_bam_output
    log:
        expand("{log_dir}/sieve_bam/{{assembly}}-{{sample}}.log", **config),
    benchmark:
        expand("{benchmark_dir}/sieve_bam/{{assembly}}-{{sample}}.benchmark.txt", **config)[0]
    message: explain_rule("sieve_bam")
    conda:
        "../envs/samtools.yaml"
    threads: 2
    params:
        minqual=f"-q {config['min_mapping_quality']}",
        atacshift=(
            lambda wildcards, input:
            f" {config['rule_dir']}/../scripts/atacshift.pl /dev/stdin {input.sizes} | "
            if config["tn5_shift"]
            else ""
        ),
        blacklist=lambda wildcards, input: f"-L {input.blacklist}",
        prim_align=f"-F 256" if config["only_primary_align"] else "",
        sizesieve=(
            lambda wildcards, input, output:
            f""" tee {output.allsizes} | awk 'substr($0,1,1)=="@" || ($9>={config['min_template_length']} && $9<={config['max_template_length']}) || ($9<=-{config['min_template_length']} && $9>=-{config['max_template_length']})' | """
            if sampledict[wildcards.sample]["layout"] == "PAIRED" and config["filter_on_size"]
            else ""
        ),
        sizesieve_touch=(
            lambda wildcards, input, output:
            f"touch {output.allsizes}"
            if sampledict[wildcards.sample]["layout"] == "SINGLE" and config["filter_on_size"]
            else ""
        ),
    shell:
        """
        samtools view -h {params.prim_align} {params.minqual} {params.blacklist} \
        {input.bam} | {params.atacshift} {params.sizesieve}
        samtools view -b > {output.final} 2> {log}
        
        # single-end reads never output allsizes so just touch the file when filtering on size
        {params.sizesieve_touch}
        """


def get_sambamba_sort_bam(wildcards):
    if sieve_bam(config):
        return rules.sieve_bam.output
    return rules.samtools_presort.output


rule sambamba_sort:
    """
    Sort the result of alignment or sieving with the sambamba sorter.
    """
    input:
        get_sambamba_sort_bam,
    output:
        temp(expand("{result_dir}/{aligner}/{{assembly}}-{{sample}}.sambamba-{{sorting}}{{sieve}}.bam", **config)),
    wildcard_constraints:
        sieve="|-sievsort",
    log:
        expand("{log_dir}/sambamba_sort/{{assembly}}-{{sample}}-sambamba_{{sorting}}{{sieve}}.log", **config),
    message: explain_rule("sambamba_sort")
    benchmark:
        expand("{benchmark_dir}/sambamba_sort/{{assembly}}-{{sample}}-{{sorting}}{{sieve}}.benchmark.txt", **config)[0]
    params:
        sort_order=lambda wildcards: "-n" if "queryname" in wildcards.sorting else "",
        memory=f"-m {config['bam_sort_mem']}G",
    threads: 2
    conda:
        "../envs/sambamba.yaml"
    resources:
        mem_gb=config["bam_sort_mem"],
    shell:
        """
        sambamba view --nthreads {threads} -f bam  {input} -o /dev/stdout  2> {log} |
        sambamba sort --nthreads {threads} {params} /dev/stdin -o {output}  2> {log}
        """


rule samtools_sort:
    """
    Sort the result of sieving with the samtools sorter.
    """
    input:
        rules.sieve_bam.output.final
    output:
        temp(expand("{result_dir}/{aligner}/{{assembly}}-{{sample}}.samtools-{{sorting}}-sievsort.bam", **config)),
    log:
        expand("{log_dir}/samtools_sort/{{assembly}}-{{sample}}-samtools_{{sorting}}.log", **config),
    benchmark:
        expand("{benchmark_dir}/samtools_sort/{{assembly}}-{{sample}}-{{sorting}}.benchmark.txt", **config)[0]
    params:
        sort_order=lambda wildcards: "-n" if wildcards.sorting == "queryname" else "",
        out_dir=f"{config['result_dir']}/{config['aligner']}",
        memory=lambda wildcards, input, output, threads: f"-m {int(1000 * round(config['bam_sort_mem']/threads, 3))}M",
    wildcard_constraints:
        sample=f"""({any_given("sample", "technical_replicates", "control")})(_allsizes)?""",
    threads: 2
    resources:
        mem_gb=config["bam_sort_mem"],
    conda:
        "../envs/samtools.yaml"
    shell:
        """
        # we set this trap to remove temp files when prematurely ending the rule
        trap "rm -f {params.out_dir}/{wildcards.assembly}-{wildcards.sample}.tmp*bam" INT;
        rm -f {params.out_dir}/{wildcards.assembly}-{wildcards.sample}.tmp*bam 2> {log}

        samtools sort {params.sort_order} -@ {threads} {params.memory} {input} -o {output} \
        -T {params.out_dir}/{wildcards.assembly}-{wildcards.sample}.tmp 2> {log}
        """


def get_bam_mark_duplicates(wildcards):
    if sieve_bam(config):
        # when alignmentsieving but not shifting we do not have to re-sort samtools-coordinate
        if wildcards.sorter == "samtools" and wildcards.sorting == "coordinate" and not config.get("tn5_shift", False):
            if config["filter_on_size"] and wildcards.sample.endswith("_allsizes"):
                return expand("{result_dir}/{aligner}/{{assembly}}-{{sample}}.samtools-coordinate-sieved.bam",**config)
            else:
                return rules.sieve_bam.output.final
        else:
            return expand("{result_dir}/{aligner}/{{assembly}}-{{sample}}.{{sorter}}-{{sorting}}-sievsort.bam", **config)

    # if we don't want to do anything get the untreated bam
    if wildcards.sorter == "samtools":
        return rules.samtools_presort.output
    return expand("{result_dir}/{aligner}/{{assembly}}-{{sample}}.{{sorter}}-{{sorting}}.bam", **config)


rule mark_duplicates:
    """
    Mark all duplicate reads in a bam file with picard MarkDuplicates
    """
    input:
        get_bam_mark_duplicates,
    output:
        bam=expand("{final_bam_dir}/{{assembly}}-{{sample}}.{{sorter}}-{{sorting}}.bam", **config),
        metrics=expand("{qc_dir}/markdup/{{assembly}}-{{sample}}.{{sorter}}-{{sorting}}.metrics.txt", **config),
    log:
        expand("{log_dir}/mark_duplicates/{{assembly}}-{{sample}}-{{sorter}}-{{sorting}}.log", **config),
    benchmark:
        expand("{benchmark_dir}/mark_duplicates/{{assembly}}-{{sample}}-{{sorter}}-{{sorting}}.benchmark.txt", **config)[0]
    message: explain_rule("mark_duplicates")
    params:
        config["markduplicates"],
    resources:
        mem_gb=8,
    conda:
        "../envs/picard.yaml"
    wildcard_constraints:
        sample=f"""({any_given("sample", "technical_replicates", "control")})(_allsizes)?""",
    shell:
        """
        # use the TMPDIR if set, and not given in the config
        if [[ ${{TMPDIR:=F}} == "F" ]] || [[ "{params}" == *TMP_DIR* ]]
        then
            tmpdir=""
        else 
            tmpdir=TMP_DIR=$TMPDIR
        fi

        picard MarkDuplicates $tmpdir {params} INPUT={input} \
        OUTPUT={output.bam} METRICS_FILE={output.metrics} > {log} 2>&1
        """


rule samtools_index:
    """
    Create an index of a bam file which can be used for e.g. visualization.
    """
    input:
        "{filepath}.bam",
    output:
        temp("{filepath}.bam.bai") if config.get("cram_no_bam", False) else "{filepath}.bam.bai",
    params:
        config["samtools_index"],
    conda:
        "../envs/samtools.yaml"
    shell:
        """
        samtools index {params} {input} {output}
        """


if config["filter_on_size"]:
    rule sam2bam:
        """
        Convert a file in sam format into a bam format (binary)
        """
        input:
            "{filepath}.sam"
        output:
            "{filepath}.bam"
        conda:
            "../envs/samtools.yaml"
        shell:
            """
            samtools view -b {input} > {output}
            """


rule bam2cram:
    """
    Convert a bam file to the more compressed cram format.
    """
    input:
        bam=rules.mark_duplicates.output.bam,
        assembly=expand("{genome_dir}/{{assembly}}/{{assembly}}.fa", **config),
    output:
        expand("{final_bam_dir}/{{assembly}}-{{sample}}.{{sorter}}-{{sorting}}.cram", **config),
    message: explain_rule("bam2cram")
    log:
        expand("{log_dir}/bam2cram/{{assembly}}-{{sample}}-{{sorter}}-{{sorting}}.log", **config),
    benchmark:
        expand("{benchmark_dir}/bam2cram/{{assembly}}-{{sample}}-{{sorter}}-{{sorting}}.benchmark.txt", **config)[0]
    params:
        threads=lambda wildcards, input, output, threads: threads - 1,
    threads: 4
    conda:
        "../envs/samtools.yaml"
    shell:
        """
        samtools view -T {input.assembly} -C {input.bam} -@ {threads} > {output} 2> {log}
        """


rule samtools_index_cram:
    """
    Generate the index for a cram file.
    """
    input:
        expand("{final_bam_dir}/{{assembly}}-{{sample}}.{{sorter}}-{{sorting}}.cram", **config),
    output:
        expand("{final_bam_dir}/{{assembly}}-{{sample}}.{{sorter}}-{{sorting}}.cram.crai", **config),
    log:
        expand("{log_dir}/samtools_index_cram/{{assembly}}-{{sample}}-{{sorter}}-{{sorting}}.log", **config),
    benchmark:
        expand("{benchmark_dir}/samtools_index_cram/{{assembly}}-{{sample}}-{{sorter}}-{{sorting}}.benchmark.txt", **config)[0]
    conda:
        "../envs/samtools.yaml"
    shell:
        """
        samtools index {input} {output} > {log} 2>&1
        """
